/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;

/**
 * Represents an arbitrary object that satisfies a particular POP application 
 * but did not have an identifier when the IdentifierHook was created.  If 
 * an identifier begins to satisfy the relevant POP application, then the 
 * IdentifierHook is notified and its getIdentifier method will return 
 * the first added identifier.
 */
public class IdentifierHook 
    implements PartialWorld.IdentifierListener {

    IdentifierHook(POP pop, List genObjs) {
	this.pop = pop;
	this.genObjs = genObjs;
    }

    public POP getPOP() {
	return pop;
    }

    public List getGeneratingObjects() {
	return genObjs;
    }

    /**
     * Returns the value of the given generating function on this object.
     */
    public Object getGenFuncValue(GeneratingFunction genFunc) {
	int index = pop.getGenFuncIndex(genFunc);
	if (index >= 0) {
	    return genObjs.get(index);
	}

	// genFunc is not used in POP that generated this obj
	return Model.NULL; 
    }

    public boolean hasIdentifier() {
	return (id != null);
    }

    /**
     * Returns the identifier associated with this hook, or null if this 
     * hook has no identifier.
     */
    public ObjectIdentifier getIdentifier() {
	return id;
    }

    public void identifierAdded(ObjectIdentifier id) {
	if (this.id == null) {
	    this.id = id;
	}
    }

    public void identifierRemoved(ObjectIdentifier id) {
	// Really this invalidates this IdentifierHook, but we do nothing.
    }

    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append("IdentifierHook for POP app (" + pop.getType() + ", ");
	for (int i = 0; i < pop.getGeneratingFunctions().size(); ++i) {
	    buf.append("(" + pop.getGeneratingFunctions().get(i) + " "
		       + genObjs.get(i) + ")");
	    if (i < pop.getGeneratingFunctions().size() - 1) {
		buf.append(", ");
	    }
	}
	buf.append("); id " + id);
	return buf.toString();
    }
    
    POP pop;
    List genObjs;
    ObjectIdentifier id = null;
}
