/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;
import java.io.PrintStream;

/**
 * Represents a type in a BLOG model.  Stores a set of potential object 
 * patterns (POPs), a set of generating functions, and a set of explicitly 
 * enumerated guaranteed objects.
 *
 * <p>We ensure that there is only one Type object for each type, so 
 * Type objects can be compared with ==.
 */
public class Type extends WrappedClass {
    /**
     * Creates a new type with the given name that is not a sub-type
     * of any other type.  It initially has no POPs, no generating
     * functions, and no enumerated guaranteed objects.
     */
    public Type(String name) {
	this.name = name;
    }

    /**
     * Creates a type with the given name that is a direct sub-type of 
     * <code>supertype</code>.  It initially has no POPs, no generating 
     * functions, and no enumerated guaranteed objects.
     *
     * <p>Note that there's no way to change the supertype of a type once 
     * it's created.  Thus, the types must be created in order from the 
     * root of the type hierarchy downward.  So it is impossible to make 
     * the supertype relation cyclic.  
     */
    public Type(String name, Type supertype) {
	this.name = name;
	this.supertype = supertype;
    }

    /**
     * Returns the name of this type.
     */
    public String getName() {
	return name;
    }

    /**
     * Returns true if this type is a subtype of the given type.  A type 
     * counts as a subtype of itself.
     */
    public boolean isSubtypeOf(Type other) {
	return ((other == this) 
		|| ((supertype != null) && supertype.isSubtypeOf(other)));
    }

    /**
     * Returns true if this is a built-in type.
     */
    public boolean isBuiltIn() {
	return isSubtypeOf(BuiltInTypes.BUILT_IN);
    }

    /**
     * Prints the type name.
     */
    public void print(PrintStream s){
	s.print(name);
    }

    /**
     * Returns the set of potential object patterns (POPs) for this type.
     *
     * @return unmodifiable Collection of POP objects
     */
    public Collection getPOPs() {
	return Collections.unmodifiableCollection(pops);
    }

    /**
     * Returns a list of generating functions for this type.  Although the 
     * order of generating functions does not affect the semantics of the 
     * model, the generating functions are always returned in the same order.  
     * This facilitates looking up the POP that uses a certain subset 
     * of the generating functions (the subset can be represented as a BitSet 
     * where bit indices correspond to the indices of generating functions 
     * in this list).  
     *
     * @return unmodifiable List of GeneratingFunction objects
     */
    public List getGeneratingFunctions() {
	return Collections.unmodifiableList(genFuncs);
    }

    /**
     * Returns the list of explicitly enumerated guaranteed objects of 
     * this type, in the order they were introduced in the model file.  
     * Or, if this is a built-in type with a finite set of guaranteed
     * objects, returns those objects in some fixed order.
     *
     * @return unmodifiable list of Object
     */
    public List getGuaranteedObjects() {
	return Collections.unmodifiableList(guaranteedObjs);
    }

    /**
     * Returns the index of the given object in this type's list of 
     * guaranteed objects.  Returns -1 if this type has infinitely many 
     * guaranteed objects (and thus does not maintain a list of them) 
     * or if the given object is not a guaranteed object of this type.
     */
    public int getGuaranteedObjIndex(Object obj) {
	if (obj instanceof EnumeratedObject) {
	    EnumeratedObject enumObj = (EnumeratedObject) obj;
	    if (enumObj.getType().equals(this)) {
		return enumObj.getIndex();
	    }
	}
	return -1;
    }

    /**
     * Returns the guaranteed object of this type that has the given 
     * index, or null if no such object exists.
     */
    public Object getGuaranteedObject(int index) {
	List guarObjs = getGuaranteedObjects();
	if ((index >= 0) && (index < guarObjs.size())) {
	    return guarObjs.get(index);
	}
	return null;
    }

    /**
     * Returns the POP that uses the given set of generating functions, 
     * or null if no such POP exists.
     *
     * @param genFuncSet a BitSet where bit i stands for the ith generating 
     *                   function in the list returned by getGeneratingFunctions
     */
    public POP getPOPWithGenFuncs(BitSet genFuncSet) {
	return (POP) genFuncsToPOP.get(genFuncSet);
    }

    /**
     * Returns the types that are strict ancestors of this type 
     * in the type graph.  The type graph contains a node for each type, 
     * with an edge from type sigma to type tau if sigma is the return 
     * type of some generating function for tau.  Note that this graph may 
     * be cyclic and may even include self-loops.  The strict ancestors of 
     * a this type are those types tau such that there is a nonzero-length 
     * directed path from tau to this type.  
     *
     * @return a Set of Type objects
     */
    public Set getStrictAncestors() {
	Set strictAncestors = new HashSet();
	
	// Find ancestors by depth-first search with repeated-node checking
	Stack stack = new Stack();
	stack.push(this);
	while (!stack.isEmpty()) {
	    Type curType = (Type) stack.pop();
	    for (Iterator iter = curType.getGeneratingFunctions().iterator(); 
		 iter.hasNext(); ) {
		Type parentType = ((Function) iter.next()).getRetType();
		if (!strictAncestors.contains(parentType)) {
		    strictAncestors.add(parentType);
		    stack.push(parentType);
		}
	    }
	}

	return strictAncestors;
    }

    /**
     * Adds the given POP to the set of POPs for this type. 
     *
     * @throws IllegalArgumentException if the given POP uses the same set 
     *                                  of generating functions as an 
     *                                  existing POP
     */
    public void addPOP(POP pop) {
	BitSet genFuncSet = pop.getGenFuncSet();
	if (getPOPWithGenFuncs(genFuncSet) != null) {
	    throw new IllegalArgumentException
		("POP " + pop + " uses same set of generating"
		 + " functions as existing POP.");
	}
	pops.add(pop);
	genFuncsToPOP.put(genFuncSet, pop);
    }

    /**
     * Adds the given function as a generating function for this type.
     */
    public void addGeneratingFunction(GeneratingFunction f) {
	genFuncs.add(f);
    }

    /**
     * Adds a new guaranteed object with the given name.  Warning: for 
     * efficiency, we do not check to make sure a guaranteed object has 
     * not already been added with the same name.  
     * 
     * @return the object added
     */
    public EnumeratedObject addGuaranteedObject(String name) {
	EnumeratedObject obj 
	    = new EnumeratedObject(this, guaranteedObjs.size(), name);
	guaranteedObjs.add(obj);
	return obj;
    }

    /**
     * Returns the number of Type objects created before this one.
     */
    public int getCreationIndex() {
	return creationIndex;
    }

    /**
     * Returns the name of this type.
     */
    public String toString() {
	return name;
    }

    private String name;
    private Type supertype;
    private int creationIndex = Type.numCreated++;

    private List pops = new ArrayList(); // of POP
    private List genFuncs = new ArrayList(); // of GeneratingFunction
    private List guaranteedObjs = new ArrayList(); // of EnumeratedObject

    private Map genFuncsToPOP = new HashMap(); // from BitSet to POP

    private static int numCreated = 0;
}
