/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog.distrib;

import blog.*;
import java.util.*;
import Jama.Matrix;
import common.Util;

/**
 * Distribution over a finite set of possible values numbered 0, ..., k-1, 
 * parameterized by a vector of probabilities pi<sub>0</sub>, ..., 
 * pi<sub>k-1</sub> that sum to one.  The probability vector can be specified 
 * when the distribution is constructed, or as an argument to the 
 * getProb and sampleVal methods.  
 */
public class Categorical extends AbstractCondProbDistrib {
    /**
     * Creates a new Categorical distribution representing the uniform 
     * distribution over the given number of values.
     */
    public Categorical(int numValues) {
	probs = new double[numValues];
	Arrays.fill(probs, 1.0 / numValues);
    }

    /**
     * Creates a new Categorical distribution with the given probability 
     * vector.  The number of values is the length of this vector.
     */
    public Categorical(double[] probs) {
	this.probs = new double[probs.length];
	for (int i = 0; i < probs.length; ++i) {
	    this.probs[i] = probs[i];
	}
    }

    /**
     * Creates a new Categorical distribution from the given parameter 
     * list.  If the parameter list is empty, then the probability vector 
     * must be passed as an argument to this CPD.  Otherwise, the parameter 
     * list should contain the probability vector as a sequence of Number 
     * objects.
     */
    public Categorical(List params) {
	if (params.isEmpty()) {
	    probs = null;
	}

	probs = new double[params.size()];
	for (int i = 0; i < params.size(); ++i) {
	    probs[i] = ((Number) params.get(i)).doubleValue();
	}
    }

    /**
     * Returns the probability of the value with the given index under 
     * this distribution.  If the given index is greater than the length 
     * of this distribution's probability vector, then the return value is 
     * zero.  
     *
     * @throws IllegalStateException    if the probability vector was not 
     *                                  specified when this distribution was 
     *                                  constructed
     *
     * @throws IllegalArgumentException if <code>index</code> is negative
     */
    public double getProb(int index) {
	if (probs == null) {
	    throw new IllegalStateException
		("Categorical distribution was constructed without a "
		 + "probability vector.");
	}

	if (index < 0) {
	    throw new IllegalArgumentException
		("Negative index passed to getProb: " + index);
	}
	if (index >= probs.length) {
	    return 0.0;
	}
	return probs[index];
    }

    /**
     * Returns the probability of the given child value conditioned on 
     * the given argument values.  If this distribution was constructed with 
     * a probability vector, then no arguments are expected.  Otherwise, 
     * there should be one argument, namely a column vector of probabilities.
     * 
     * @return the probability associated with <code>childValue</code>'s 
     *         index, or 0 if <code>childValue</code> does not have an index 
     *         or its index is beyond the end of the probability vector
     *
     * @throws IllegalArgumentException  if the probability vector was 
     *                                   specified at construction and 
     *                                   <code>args</code> is non-empty, or 
     *                                   if the probability vector was not 
     *                                   specified and <code>args</code> is 
     *                                   empty
     */
    public double getProb(List args, Object childValue) {
	ensureProbsInited(args);

	int index = Model.getObjectIndex(childValue);
	if ((index == -1) || (index >= probs.length)) {
	    return 0;
	}
	return probs[index];
    }

    /**
     * Returns an index sampled from this distribution.
     *
     * @throws IllegalStateException    if the probability vector was not 
     *                                  specified when this distribution was 
     *                                  constructed
     */
    public int sampleVal() {
	if (probs == null) {
	    throw new IllegalStateException
		("Categorical distribution was constructed without a "
		 + "probability vector.");
	}

	return Util.sampleWithProbs(probs);
    }

    /**
     * Returns a value sampled from the given child type according to this 
     * distribution.  If this distribution was constructed with 
     * a probability vector, then no arguments are expected.  Otherwise, 
     * there should be one argument, namely a column vector of probabilities.
     * 
     * @return a guaranteed object of the given type whose index is sampled 
     *         according to this distribution, or Model.NULL if the index 
     *         sampled is beyond the end of the guaranteed object list for 
     *         the given type
     *
     * @throws IllegalArgumentException  if the probability vector was 
     *                                   specified at construction and 
     *                                   <code>args</code> is non-empty, or 
     *                                   if the probability vector was not 
     *                                   specified and <code>args</code> is 
     *                                   empty
     */
    public Object sampleVal(List args, Type childType) {
	ensureProbsInited(args);

	int index = Util.sampleWithProbs(probs);
	Object value = childType.getGuaranteedObject(index);
	if (value == null) {
	    // make list so we can print the probabilities easily
	    List probList = new ArrayList();
	    for (int i = 0; i < probs.length; ++i) {
		probList.add(new Double(probs[i]));
	    }
	    System.err.println("Warning: distribution does not sum to 1 over "
			       + "the guaranteed objects of type " + childType
			       + ": " + probList);
	    value = Model.NULL;
	}
	return value;
    }

    private void ensureProbsInited(List args) {
	if (probs == null) {
	    if (args.isEmpty()) {
		throw new IllegalArgumentException
		    ("Arguments to Categorical CPD should consist of a "
		     + "probability vector, since the probability vector "
		     + "was not specified as a CPD parameter.");
	    }
	    
	    if ((!(args.get(0) instanceof Matrix))
		    || (((Matrix) args.get(0)).getColumnDimension() != 1)) {
		throw new IllegalArgumentException
		    ("Argument to Categorical CPD should be a column "
		     + "vector of probabilities, not: " + args.get(0));
	    }

	    Matrix m = (Matrix) args.get(0);
	    probs = new double[m.getRowDimension()];
	    for (int i = 0; i < probs.length; ++i) {
		probs[i] = m.get(i, 0);
	    }
	} else if (!args.isEmpty()) {
	    throw new IllegalArgumentException
		("Categorical CPD expects no arguments (probabilities were "
		 + "specified as CPD parameters).");
	}
    }

    private double[] probs;
}
